# Copyright 2015 Spotify AB. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import re

from utils import junos_views
from base import NetworkDriver

from jnpr.junos import Device
from jnpr.junos.utils.config import Config
from jnpr.junos.exception import ConfigLoadError
from exceptions import ReplaceConfigException, MergeConfigException

from utils import string_parsers


class JunOSDriver(NetworkDriver):

    def __init__(self, hostname, username, password, timeout=60):
        self.hostname = hostname
        self.username = username
        self.password = password
        self.timeout = timeout
        self.device = Device(hostname, user=username, password=password)
        self.config_replace = False

    def open(self):
        self.device.open()
        self.device.timeout = self.timeout
        self.device.bind(cu=Config)
        self.device.cu.lock()

    def close(self):
        self.device.cu.unlock()
        self.device.close()

    def _load_candidate(self, filename, config, overwrite):
        if filename is None:
            configuration = config
        else:
            with open(filename) as f:
                configuration = f.read()

        try:
            self.device.cu.load(configuration, format='text', overwrite=overwrite)
        except ConfigLoadError as e:
            if self.config_replace:
                raise ReplaceConfigException(e.message)
            else:
                raise MergeConfigException(e.message)

    def load_replace_candidate(self, filename=None, config=None):
        self.config_replace = True
        self._load_candidate(filename, config, True)

    def load_merge_candidate(self, filename=None, config=None):
        self.config_replace = False
        self._load_candidate(filename, config, False)

    def compare_config(self):
        diff = self.device.cu.diff()

        if diff is None:
            return ''
        else:
            return diff.strip()

    def commit_config(self):
        self.device.cu.commit()

    def discard_config(self):
        self.device.cu.rollback(rb_id=0)

    def rollback(self):
        self.device.cu.rollback(rb_id=1)
        self.commit_config()

    def get_facts(self):

        output = self.device.facts

        uptime = 0
        if 'RE0' in output:
            uptime = output['RE0']['up_time']

        interfaces = junos_views.junos_iface_table(self.device)
        interfaces.get()
        interface_list = interfaces.keys()

        return {
            'vendor': u'Juniper',
            'model': unicode(output['model']),
            'serial_number': unicode(output['serialnumber']),
            'os_version': unicode(output['version']),
            'hostname': unicode(output['hostname']),
            'fqdn': unicode(output['fqdn']),
            'uptime': string_parsers.convert_uptime_string_seconds(uptime),
            'interface_list': interface_list
        }

    def get_interfaces(self):

        # init result dict
        result = {}

        interfaces = junos_views.junos_iface_table(self.device)
        interfaces.get()

        # convert all the tuples to our pre-defined dict structure
        for iface in interfaces.keys():
            result[iface] = {
                'is_up': interfaces[iface]['is_up'],
                'is_enabled': interfaces[iface]['is_enabled'],
                'description': interfaces[iface]['description'] or u'',
                'last_flapped': interfaces[iface]['last_flapped'] or -1,
                'mac_address': unicode(interfaces[iface]['mac_address'])
            }
            result[iface]['last_flapped'] = float(result[iface]['last_flapped'])

            match = re.search(r'\d+', interfaces[iface]['speed'] or '')
            if match is not None:
                result[iface]['speed'] = int(match.group(0))
            else:
                result[iface]['speed'] = -1

        return result

    # def get_bgp_neighbors(self):
    #
    #     # init result dict
    #     result = {}
    #
    #     instances = junos_views.junos_route_instance_table(self.device)
    #     instances.get()
    #
    #     for vrf in instances.keys():
    #         if not vrf.startswith('__'):
    #
    #             # init result dict for this vrf
    #             result[vrf] = {
    #                 'peers': {},
    #                 'router_id': None,
    #                 'local_as': None,
    #             }
    #
    #             # fetch sessions for vrf
    #             bgp = junos_views.junos_bgp_table(self.device)
    #             bgp.get(instance=vrf)
    #
    #             # assemble result dict
    #             bgp_result = {}
    #             [bgp_result.update({neigh:dict(bgp[neigh])}) for neigh in bgp.keys()]
    #             result[vrf]['peers'] = bgp_result
    #
    #     return result

    def get_lldp_neighbors(self):
        lldp = junos_views.junos_lldp_table(self.device)
        lldp.get()

        result = lldp.items()

        neighbors = dict()
        for neigh in result:
            if neigh[0] not in neighbors.keys():
                neighbors[neigh[0]] = list()
            neighbors[neigh[0]].append({ x[0]: unicode(x[1]) for x in neigh[1]})

        return neighbors



#
# Copyright 2015 Kamil Derynski, Opera Software ASA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

from base import NetworkDriver
from exceptions import ReplaceConfigException, MergeConfigException
from bnclient import bnclient
import difflib
import sys
from StringIO import StringIO


class IBMDriver(NetworkDriver):

    def __init__(self, hostname, username, password, timeout=60):
        self.hostname = hostname
        self.username = username
        self.password = password
        self.timeout = timeout
        self.argv = ['', '-u', username, '-p', password, hostname]
        self.bnc = bnclient.bnclient(self.argv)
        self.config_replace = False
        self.config = {"running": "", "candidate": "", "rollback": ""}
        self.filename_running = "/tmp/" + hostname + "-running.conf"
        self.filename_candidate = None
        self.filename_rollback = "/tmp/" + hostname + "-rollback.conf"
        self.error = StringIO()

    def str2argv(self, str=''):
        return str.split(' ')

    def open(self):
        self.bnc.connect()
        self.bnc.sendhello()
        self._get_config(self.filename_rollback)

    def close(self):
        self.bnc.close()

    def _bnc_cmd(self, file, operation, error_option):
        cmd = "-o edit-config -t running -f "
        cmd += file
        cmd += " -d " + operation
        cmd += " -r " + error_option
        return cmd

    def _write_memory(self):
        cmd = "-o copy-config -t startup -s running"
        return cmd

    def _send_rpc(self, cmd):
        self.error = StringIO()
        old_stdout = sys.stdout
        sys.stdout = self.error
        self.bnc.sendrpc(self.str2argv(cmd))
        sys.stdout = old_stdout

    def _get_config(self, filename):
        self.bnc.sendrpc(self.str2argv("-o get -f " + filename))

    def _load_config(self, filename, config):
        self._get_config(self.filename_rollback)
        if filename is None:
            configuration = config
        else:
            with open(filename) as f:
                configuration = f.read()
                self.config['candidate'] = configuration
                self.filename_candidate = filename
                f.close()

    def load_replace_candidate(self, filename=None, config=None):
        self.config_replace = True
        self._load_config(filename, config)

    def load_merge_candidate(self, filename=None, config=None):
        self.config_replace = False
        self._load_config(filename, config)

    def compare_config(self):
        result = ''
        if self.config_replace:
            self._get_config(self.filename_running)
            with open(self.filename_running, 'r') as running:
                with open(self.filename_candidate, 'r') as candidate:
                    diff = difflib.unified_diff(
                        running.readlines(),
                        candidate.readlines(),
                        fromfile='running',
                        tofile='candidate',
                    )
            for line in diff:
                for prefix in ('---', '+++', '@@'):
                    if line.startswith(prefix):
                        break
                else:
                    result += line
        else:
            result = self.config['candidate']
        return str(result).strip()

    def _commit_replace(self):
        cmd = self._bnc_cmd(self.filename_candidate, "replace", "continue-on-error")
        self._send_rpc(cmd)
        if self.error.getvalue():
            self.rollback()
            raise ReplaceConfigException(self.error.getvalue())

    def _commit_merge(self):
        cmd = self._bnc_cmd(self.filename_candidate, "merge", "rollback-on-error")
        self._send_rpc(cmd)
        if self.error.getvalue():
            self.discard_config()
            raise MergeConfigException(self.error.getvalue())

    def commit_config(self):
        if self.config_replace:
            self._commit_replace()
        else:
            self._commit_merge()
        cmd = self._write_memory()
        self._send_rpc(cmd)

    def discard_config(self):
        self.filename_candidate = self.filename_running
        self.config['candidate'] = self.config['running']

    def rollback(self):
        cmd = self._bnc_cmd(self.filename_rollback, "replace", "rollback-on-error")
        self._send_rpc(cmd)
        cmd = self._write_memory()
        self._send_rpc(cmd)



# Copyright 2015 Spotify AB. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import pyeapi

from base import NetworkDriver

from exceptions import MergeConfigException, ReplaceConfigException, SessionLockedException

from datetime import datetime
import time

from utils import string_parsers


class EOSDriver(NetworkDriver):

    def __init__(self, hostname, username, password, timeout=60):
        self.device = None
        self.hostname = hostname
        self.username = username
        self.password = password
        self.timeout = timeout
        self.config_session = None

    def open(self):
        connection = pyeapi.client.connect(
            transport='https',
            host=self.hostname,
            username=self.username,
            password=self.password,
            port=443
        )
        self.device = pyeapi.client.Node(connection)

    def close(self):
        self.discard_config()

    def _load_config(self, filename=None, config=None, replace=True):
        if self.config_session is not None:
            raise SessionLockedException('Session is already in use by napalm')
        else:
            self.config_session = 'napalm_{}'.format(datetime.now().microsecond)

        commands = list()
        commands.append('configure session {}'.format(self.config_session))

        if replace:
            commands.append('rollback clean-config')

        if filename is not None:
            with open(filename, 'r') as f:
                lines = f.readlines()
        else:
            if isinstance(config, list):
                lines = config
            else:
                lines = config.splitlines()

        for line in lines:
            line = line.strip()
            if line == '':
                continue
            if line.startswith('!'):
                continue
            commands.append(line)

        try:
            self.device.run_commands(commands)
        except pyeapi.eapilib.CommandError as e:
            self.discard_config()

            if replace:
                raise ReplaceConfigException(e.message)
            else:
                raise MergeConfigException(e.message)

    def load_replace_candidate(self, filename=None, config=None):
        self._load_config(filename, config, True)

    def load_merge_candidate(self, filename=None, config=None):
        self._load_config(filename, config, False)

    def compare_config(self):
        if self.config_session is None:
            return ''
        else:
            commands = ['show session-config named %s diffs' % self.config_session]
            result = self.device.run_commands(commands, encoding='text')[0]['output']

            result = '\n'.join(result.splitlines()[2:])

            return result.strip()

    def commit_config(self):
        commands = list()
        commands.append('copy startup-config flash:rollback-0')
        commands.append('configure session {}'.format(self.config_session))
        commands.append('commit')
        commands.append('write memory')

        self.device.run_commands(commands)
        self.config_session = None

    def discard_config(self):
        if self.config_session is not None:
            commands = list()
            commands.append('configure session {}'.format(self.config_session))
            commands.append('abort')
            self.device.run_commands(commands)
            self.config_session = None

    def rollback(self):
        commands = list()
        commands.append('configure replace flash:rollback-0')
        commands.append('write memory')
        self.device.run_commands(commands)

    def get_facts(self):
        commands = list()
        commands.append('show version')
        commands.append('show hostname')
        commands.append('show interfaces status')

        result = self.device.run_commands(commands)

        version = result[0]
        hostname = result[1]
        interfaces_dict = result[2]['interfaceStatuses']

        uptime = time.time() - version['bootupTimestamp']

        interfaces = [i for i in interfaces_dict.keys() if '.' not in i]
        interfaces = string_parsers.sorted_nicely(interfaces)

        return {
            'hostname': hostname['hostname'],
            'fqdn': hostname['fqdn'],
            'vendor': u'Arista',
            'model': version['modelName'],
            'serial_number': version['serialNumber'],
            'os_version': version['internalVersion'],
            'uptime': int(uptime),
            'interface_list': interfaces,
        }

    def get_interfaces(self):
        commands = list()
        commands.append('show interfaces')
        output = self.device.run_commands(commands)[0]

        interfaces = dict()

        for interface, values in output['interfaces'].iteritems():
            interfaces[interface] = dict()

            if values['lineProtocolStatus'] == 'up':
                interfaces[interface]['is_up'] = True
                interfaces[interface]['is_enabled'] = True
            else:
                interfaces[interface]['is_up'] = False
                if values['interfaceStatus'] == 'disabled':
                    interfaces[interface]['is_enabled'] = False
                else:
                    interfaces[interface]['is_enabled'] = True

            interfaces[interface]['description'] = values['description']

            interfaces[interface]['last_flapped'] = values.pop('lastStatusChangeTimestamp', None)

            interfaces[interface]['speed'] = values['bandwidth']
            interfaces[interface]['mac_address'] = values.pop('physicalAddress', None)

        return interfaces

    def get_lldp_neighbors(self):
        commands = list()
        commands.append('show lldp neighbors')
        output = self.device.run_commands(commands)[0]['lldpNeighbors']

        lldp = dict()

        for n in output:
            if n['port'] not in lldp.keys():
                lldp[n['port']] = list()

            lldp[n['port']].append(
                {
                    'hostname': n['neighborDevice'],
                    'port': n['neighborPort'],
                }
            )

        return lldp

    # def get_bgp_neighbors(self):
    #     bgp_neighbors = dict()
    #
    #     for vrf, vrf_data in self.device.show_ip_bgp_summary_vrf_all()['vrfs'].iteritems():
    #         bgp_neighbors[vrf] = dict()
    #         bgp_neighbors[vrf]['router_id'] = vrf_data['routerId']
    #         bgp_neighbors[vrf]['local_as'] = vrf_data['asn']
    #         bgp_neighbors[vrf]['peers'] = dict()
    #
    #         for n, n_data in vrf_data['peers'].iteritems():
    #             bgp_neighbors[vrf]['peers'][n] = dict()
    #
    #             if n_data['peerState'] == 'Established':
    #                 bgp_neighbors[vrf]['peers'][n]['is_up'] = True
    #                 bgp_neighbors[vrf]['peers'][n]['is_enabled'] = True
    #             else:
    #                 bgp_neighbors[vrf]['peers'][n]['is_up'] = False
    #
    #                 reason = n_data.pop('peerStateIdleReason', None)
    #                 if reason == 'Admin':
    #                     bgp_neighbors[vrf]['peers'][n]['is_enabled'] = False
    #                 else:
    #                     bgp_neighbors[vrf]['peers'][n]['is_enabled'] = True
    #
    #             bgp_neighbors[vrf]['peers'][n]['remote_as'] = n_data['asn']
    #             bgp_neighbors[vrf]['peers'][n]['uptime'] = n_data['upDownTime']
    #
    #             raw_data = eval(
    #                 'self.device.show_ip_bgp_neighbors_vrf_{}(format="text", pipe="section {}")'.format(vrf, n)
    #             )['output']
    #
    #             n_data_full =  colon_separated_string_to_dict(raw_data)
    #             sent, rcvd = n_data_full['IPv4 Unicast'].split()
    #
    #             bgp_neighbors[vrf]['peers'][n]['received_prefixes'] = int(rcvd)
    #             bgp_neighbors[vrf]['peers'][n]['sent_prefixes'] = int(sent)
    #             bgp_neighbors[vrf]['peers'][n]['accepted_prefixes'] = n_data['prefixAccepted']
    #
    #             bgp_neighbors[vrf]['peers'][n]['description'] = n_data_full.pop('Description', '')
    #
    #     return bgp_neighbors



# Copyright 2015 Spotify AB. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from pyFG.fortios import FortiOS, FortiConfig
from pyFG.exceptions import FailedCommit, CommandExecutionException

from base import NetworkDriver
from exceptions import ReplaceConfigException, MergeConfigException

from utils.string_parsers import colon_separated_string_to_dict, convert_uptime_string_seconds


def execute_get(device, cmd, separator=':', auto=False):
    output = device.execute_command(cmd)

    if auto:
        if ':' in output[0]:
            separator=':'
        elif '\t' in output[0]:
            separator='\t'
        else:
            raise Exception('Unknown separator for block:\n{}'.format(output))

    return colon_separated_string_to_dict('\n'.join(output), separator)


class FortiOSDriver(NetworkDriver):

    def __init__(self, hostname, username, password, timeout=60):
        self.hostname = hostname
        self.username = username
        self.password = password
        self.device = FortiOS(hostname, username=username, password=password, timeout=timeout)
        self.config_replace = False

    def open(self):
        self.device.open()

    def close(self):
        self.device.close()

    def _load_config(self, filename, config):
        if filename is None:
            configuration = config
        else:
            with open(filename) as f:
                configuration = f.read()

        self.device.load_config(in_candidate=True, config_text=configuration)

    def load_replace_candidate(self, filename=None, config=None):
        self.config_replace = True

        self.device.candidate_config = FortiConfig('candidate')
        self.device.running_config = FortiConfig('running')

        self._load_config(filename, config)

        self.device.load_config(empty_candidate=True)

    def load_merge_candidate(self, filename=None, config=None):
        self.config_replace = False

        self.device.candidate_config = FortiConfig('candidate')
        self.device.running_config = FortiConfig('running')

        self._load_config(filename, config)

        for block in self.device.candidate_config.get_block_names():
            try:
                self.device.load_config(path=block, empty_candidate=True)
            except CommandExecutionException as e:
                raise MergeConfigException(e.message)

    def compare_config(self):
        return self.device.compare_config()

    def commit_config(self):
        try:
            self.device.execute_command('execute backup config flash commit_with_napalm')
            self.device.commit()
            self.discard_config()
        except FailedCommit as e:
            if self.config_replace:
                raise ReplaceConfigException(e.message)
            else:
                raise MergeConfigException(e.message)

    def discard_config(self):
        self.device.candidate_config = FortiConfig('candidate')
        self.device.load_config(in_candidate=True)

    def rollback(self):
        output = self.device.execute_command('fnsysctl ls -l data2/config')
        rollback_file = output[-2].split()[-1]
        rollback_config = self.device.execute_command('fnsysctl cat data2/config/%s' % rollback_file)

        self.device.load_config(empty_candidate=True)
        self.load_replace_candidate(config=rollback_config)
        self.device.candidate_config['vpn certificate local']['Fortinet_CA_SSLProxy'].del_param('private-key')
        self.device.candidate_config['vpn certificate local']['Fortinet_CA_SSLProxy'].del_param('certificate')
        self.device.candidate_config['vpn certificate local']['Fortinet_SSLProxy'].del_param('private-key')
        self.device.candidate_config['vpn certificate local']['Fortinet_SSLProxy'].del_param('certificate')
        self.device.commit()

    def get_facts(self):
        system_status = execute_get(self.device, 'get system status')
        performance_status = execute_get(self.device, 'get system performance status')

        interfaces = execute_get(self.device, 'get system interface | grep ==')
        interface_list = [x.split()[2] for x in interfaces.keys()]

        domain = execute_get(self.device, 'get system dns | grep domain')['domain']

        return {
            'vendor': 'Fortigate',
            'os_version': system_status['Version'].split(',')[0].split()[1],
            'uptime': convert_uptime_string_seconds(performance_status['Uptime']),
            'serial_number': system_status['Serial-Number'],
            'model': system_status['Version'].split(',')[0].split()[0],
            'hostname': system_status['Hostname'],
            'fqdn': '{}.{}'.format(system_status['Hostname'], domain),
            'interface_list': interface_list
        }

    @staticmethod
    def _get_tab_separated_interfaces(output):
        interface_statistics = {
            'is_up': ('up' in output['State'] and 'up' or 'down'),
            'speed': output['Speed'],
            'mac_adddress': output['Current_HWaddr']
        }
        return interface_statistics

    @staticmethod
    def _get_unsupported_interfaces():
        return {
            'is_up': None,
            'is_enabled': None,
            'description': None,
            'last_flapped': None,
            'mode': None,
            'speed': None,
            'mac_address': None
        }

    def get_interfaces(self):
        interface_statistics = dict()
        self.device.load_config('system interface')
        for iface in self.get_facts()['interface_list']:
            try:
                hw_output = execute_get(self.device, 'diagnose hardware deviceinfo nic {}'.format(iface), auto=True)
                ifs = self._get_tab_separated_interfaces(hw_output)

                ifs['is_enabled'] = self.device.running_config['system interface'][iface].get_param('status') != 'down'
                ifs['description'] = self.device.running_config['system interface'][iface].get_param('description')
                ifs['last_flapped'] = None
                #ifs['mode'] = 'routed'
                ifs['last_flapped'] = None
                interface_statistics[iface] = ifs
            except CommandExecutionException:
                interface_statistics[iface] = self._get_unsupported_interfaces()

        return interface_statistics

    # def get_bgp_neighbors(self):
    #     bgp_sum = self.device.execute_command('get router info bgp sum')
    #     re_neigh = re.compile("^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+")
    #     neighbors = {n.split()[0]: n.split()[1:] for n in bgp_sum if re.match(re_neigh, n)}
    #
    #     peers = dict()
    #
    #     self.device.load_config('router bgp')
    #
    #     for neighbor, parameters in neighbors.iteritems():
    #         neigh_conf = self.device.running_config['router bgp']['neighbor']['{}'.format(neighbor)]
    #         peers[neighbor] = dict()
    #         peers[neighbor]['remote_as'] = int(neigh_conf.get_param('remote-as'))
    #         peers[neighbor]['is_enabled'] = neigh_conf.get_param('shutdown') != 'enable' or False
    #         peers[neighbor]['is_up'] = 'never' != parameters[7] or False
    #         peers[neighbor]['uptime'] = convert_uptime_string_seconds(parameters[7])
    #         peers[neighbor]['accepted_prefixes'] = int(self.device.execute_command('get router info bgp neighbor {} | grep "accepted prefixes"'.format(neighbor))[0].split()[0])
    #         peers[neighbor]['sent_prefixes'] = int(self.device.execute_command('get router info bgp neighbor {} | grep "announced prefixes"'.format(neighbor))[0].split()[0])
    #
    #         received = self.device.execute_command('get router info bgp neighbors {} received-routes | grep prefixes'.format(neighbor))[0].split()
    #         if len(received) > 0:
    #             peers[neighbor]['received_prefixes'] = int(received[-1])
    #         else:
    #             # Soft-reconfig is not enabled
    #             peers[neighbor]['received_prefixes'] = 0
    #
    #     return {
    #         'default': {
    #             'local_as': int(bgp_sum[0].split()[7]),
    #             'router_id': bgp_sum[0].split()[3],
    #             'peers': peers
    #         }
    #     }

    def get_lldp_neighbors(self):
        return {}



# Copyright 2015 Spotify AB. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.


class NetworkDriver:

    def __init__(self, hostname, username, password, timeout):
        """
        This is the base class you have to inherit from when writing your own Network Driver to manage any device. You
        will, in addition, have to override all the methods specified on this class. Make sure you follow the guidelines
        for every method and that you return the correct data.

        :param hostname: (str) IP or FQDN of the device you want to connect to.
        :param username: (str) Username you want to use
        :param password: (str) Password
        :param timeout: (int) Time in seconds to wait for the device to respond.
        :return:
        """
        raise NotImplementedError

    def open(self):
        """
        Opens a connection to the device.
        """
        raise NotImplementedError

    def close(self):
        """
        Closes the connection to the device.
        """
        raise NotImplementedError

    def load_replace_candidate(self, filename=None, config=None):
        """
        Populates the candidate configuration. You can populate it from a file or from a string. If you send both a
        filename and a string containing the configuration, the file takes precedence.

        If you use this method the existing configuration will be replaced entirely by the candidate configuration once
        you commit the changes. This method will not change the configuration by itself.

        :param filename: Path to the file containing the desired configuration. By default is None.
        :param config: String containing the desired configuration.
        :raise ReplaceConfigException: If there is an error on the configuration sent.
        """
        raise NotImplementedError

    def load_merge_candidate(self, filename=None, config=None):
        """
        Populates the candidate configuration. You can populate it from a file or from a string. If you send both a
        filename and a string containing the configuration, the file takes precedence.

        If you use this method the existing configuration will be merged with the candidate configuration once
        you commit the changes. This method will not change the configuration by itself.

        :param filename: Path to the file containing the desired configuration. By default is None.
        :param config: String containing the desired configuration.
        :raise MergeConfigException: If there is an error on the configuration sent.
        """
        raise NotImplementedError

    def compare_config(self):
        """
        :return: A string showing the difference between the running configuration and the candidate configuration. The\
        running_config is loaded automatically just before doing the comparison so there is no need for you to do it.
        """
        raise NotImplementedError

    def commit_config(self):
        """
        Commits the changes requested by the method load_replace_candidate or load_merge_candidate.
        """
        raise NotImplementedError

    def discard_config(self):
        """
        Discards the configuration loaded into the candidate.
        """
        raise NotImplementedError

    def rollback(self):
        """
        If changes were made, revert changes to the original state.
        """
        raise NotImplementedError

    def get_facts(self):
        """
        Returns a dictionary containing the following information:
         * uptime - Uptime of the device in seconds.
         * vendor - Manufacturer of the device.
         * model - Device model.
         * hostname - Hostname of the device
         * fqdn - Fqdn of the device
         * os_version - String with the OS version running on the device.
         * serial_number - Serial number of the device
         * interface_list - List of the interfaces of the device

        For example::

            {
            'uptime': 151005.57332897186,
            'vendor': u'Arista',
            'os_version': u'4.14.3-2329074.gaatlantarel',
            'serial_number': u'SN0123A34AS',
            'model': u'vEOS',
            'hostname': u'eos-router',
            'fqdn': u'eos-router',
            'interface_list': [u'Ethernet2', u'Management1', u'Ethernet1', u'Ethernet3']
            }

        """


        raise NotImplementedError

    def get_interfaces(self):
        """
        Returns a dictionary of dictionaries. The keys for the first dictionary will be the interfaces in the devices.\
        The inner dictionary will containing the following data for each interface:
         * is_up (True/False)
         * is_enabled (True/False)
         * description (string)
         * last_flapped (int in seconds)
         * speed (int in Mbit)
         * mac_address (string)

        For example::

            {
            u'Management1':
                {
                'is_up': False,
                'is_enabled': False,
                'description': u'',
                'last_flapped': -1,
                'speed': 1000,
                'mac_address': u'dead:beef:dead',
                },
            u'Ethernet1':
                {
                'is_up': True,
                'is_enabled': True,
                'description': u'foo',
                'last_flapped': 1429978575.1554043,
                'speed': 1000,
                'mac_address': u'beef:dead:beef',
                },
            u'Ethernet2':
                {
                'is_up': True,
                'is_enabled': True,
                'description': u'bla',
                'last_flapped': 1429978575.1555667,
                'speed': 1000,
                'mac_address': u'beef:beef:beef',
                },
            u'Ethernet3':
                {
                'is_up': False,
                'is_enabled': True,
                'description': u'bar',
                'last_flapped': -1,
                'speed': 1000,
                'mac_address': u'dead:dead:dead',
                }
            }
        """
        raise NotImplementedError

    def get_lldp_neighbors(self):
        """
        Returns a dictionary where the keys are local ports and the value is a list of dictionaries with the following \
        information:
            * hostname
            * port

        For example::

            {
            u'Ethernet2':
                [
                    {
                    'hostname': u'junos-unittest',
                    'port': u'520',
                    }
                ],
            u'Ethernet3':
                [
                    {
                    'hostname': u'junos-unittest',
                    'port': u'522',
                    }
                ],
            u'Ethernet1':
                [
                    {
                    'hostname': u'junos-unittest',
                    'port': u'519',
                    },
                    {
                    'hostname': u'ios-xrv-unittest',
                    'port': u'Gi0/0/0/0',
                    }
                ],
            u'Management1':
                [
                    {
                    'hostname': u'junos-unittest',
                    'port': u'508',
                    }
                ]
            }
        """
        raise NotImplementedError

    # def get_bgp_neighbors(self):
    #     """
    #     Returns a dictionary of dictionaries. The keys for the first dictionary will be the vrf (global if no vrf).\
    #     The inner dictionary will containg the following data for each vrf:
    #      * local_as (int)
    #      * router_id
    #      * peers - another dictionary of dictionaries. Outer keys are the IPs of the neighbors. The inner keys are:
    #        * is_up (True/False)
    #        * is_enabled (True/False)
    #        * remote_as (int)
    #        * description (string)
    #        * uptime (int in seconds)
    #        * received_prefixes (int)
    #        * accepted_prefixes (int)
    #        * sent_prefixes (int)
    #
    #     For example:
    #
    #         {
    #         u'default':
    #             {
    #             'router_id': u'192.168.0.1',
    #             'local_as': 65000,
    #             'peers':
    #                 {
    #                 u'10.0.0.11':
    #                     {
    #                     'is_up': True,
    #                     'is_enabled': True,
    #                     'uptime': 1429978587.950959,
    #                     'description': u'',
    #                     'received_prefixes': 2,
    #                     'sent_prefixes': 3,
    #                     'accepted_prefixes': 0,
    #                     'remote_as': 65001
    #                     },
    #                 u'1.1.1.1':
    #                     {
    #                     'is_up': False,
    #                     'is_enabled': False,
    #                     'uptime': -1,
    #                     'description': u'',
    #                     'received_prefixes': 0,
    #                     'sent_prefixes': 0,
    #                     'accepted_prefixes': 0,
    #                     'remote_as': 1
    #                     }
    #                 }
    #             },
    #         u'vrfA':
    #             {
    #             'router_id': u'10.0.1.10',
    #             'local_as': 65010,
    #             'peers':
    #                 {
    #                 u'10.0.1.12':
    #                     {
    #                     'is_up': False,
    #                     'is_enabled': True,
    #                     'uptime': -1,
    #                     'description': u'',
    #                     'received_prefixes': 0,
    #                     'sent_prefixes': 0,
    #                     'accepted_prefixes': 0,
    #                     'remote_as': 65012
    #                     },
    #                 u'10.0.1.13':
    #                     {
    #                     'is_up': False,
    #                     'is_enabled': True,
    #                     'uptime': -1,
    #                     'description': u'',
    #                     'received_prefixes': 0,
    #                     'sent_prefixes': 0,
    #                     'accepted_prefixes': 0,
    #                     'remote_as': 65013
    #                     }
    #                 }
    #             }
    #         }
    #
    #
    #     """
    #     raise NotImplementedError
